'*******************************************************************************
' *   This file is part of NRtfTree Library.
' *
' *   NRtfTree Library is free software; you can redistribute it and/or modify
' *   it under the terms of the GNU Lesser General Public License as published by
' *   the Free Software Foundation; either version 3 of the License, or
' *   (at your option) any later version.
' *
' *   NRtfTree Library is distributed in the hope that it will be useful,
' *   but WITHOUT ANY WARRANTY; without even the implied warranty of
' *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' *   GNU Lesser General Public License for more details.
' *
' *   You should have received a copy of the GNU Lesser General Public License
' *   along with this program. If not, see <http://www.gnu.org/licenses/>.
' *******************************************************************************

'*******************************************************************************
' * Library:		NRtfTree
' * Version:     v0.3.0
' * Date:		01/05/2009
' * Copyright:   2006-2009 Salvador Gomez
' * E-mail:      sgoliver.net@gmail.com
' * Home Page:	http://www.sgoliver.net
' * SF Project:	http://nrtftree.sourceforge.net
' *				http://sourceforge.net/projects/nrtftree
' * Class:		Rtf2Html
' * Description:	Traduccin de documentos RTF a formato HTML.
' * Notes:       Contribucin de Francisco Javier Marn (http://www.xuletas.es/).
' * Contributed by: Terence L. Wallace
' *******************************************************************************


Imports Microsoft.VisualBasic
Imports System
Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.IO
Imports System.Text
Imports System.Text.RegularExpressions
Imports Net.Sgoliver.NRtfTree.Core
Imports Net.Sgoliver.NRtfTree.Util

Namespace Net.Sgoliver.NRtfTree
	Namespace Demo
		''' <summary>
		''' Conversor de documentos RTF a formato HTML.
		''' </summary>
		Public Class Rtf2Html
			#Region "Atributos Privados"

			''' <summary>
			''' StringBuilder empleado para escribir el cdigo HTML resultado
			''' </summary>
			Private _builder As StringBuilder

			''' <summary>
			''' Formato de texto que se est leyendo
			''' </summary>
			Private _currentFormat As Format

			''' <summary>
			''' Formato de texto ya escrito en el cdigo HTML
			''' </summary>
			Private _htmlFormat As Format

			''' <summary>
			''' Tabla de colores RTF
			''' </summary>
			Private _colorTable As RtfColorTable

			''' <summary>
			''' Tabla de fuentes RTF
			''' </summary>
			Private _fontTable As RtfFontTable

			'Propiedades

			Private _autoParagraph As Boolean
			Private _ignoreFontNames As Boolean
			Private _escapeHtmlEntities As Boolean
			Private _defaultFontName As String
			Private _defaultFontSize As Integer
			Private _incrustImages As Boolean
			Private _imagePath As String
			Private _imageFormat As ImageFormat

			#End Region

			#Region "Constructores"

			Public Sub New()
				AutoParagraph = True
				IgnoreFontNames = False
				DefaultFontSize = 10
				EscapeHtmlEntities = True
				DefaultFontName = "Times New Roman"

				ImageFormat = ImageFormat.Jpeg
				IncrustImages = True
				ImagePath = ""
			End Sub

			#End Region

			#Region "Propiedades"

			''' <summary>
			''' Obtiene o establece un valor que indica si se crearn parrafos automticamente
			''' </summary>
			Public Property AutoParagraph() As Boolean
				Get
					Return _autoParagraph
				End Get
				Set(ByVal value As Boolean)
					_autoParagraph = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece un valor que indica si se ignorarn los nombres de las fuentes.
			''' Establecer este valor como false generar un archivo HTML sin propiedad CSS font-face
			''' </summary>
			Public Property IgnoreFontNames() As Boolean
				Get
					Return _ignoreFontNames
				End Get
				Set(ByVal value As Boolean)
					_ignoreFontNames = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece un valor que indica si se reemplazarn las entidades HTML
			''' del texto
			''' </summary>
			Public Property EscapeHtmlEntities() As Boolean
				Get
					Return _escapeHtmlEntities
				End Get
				Set(ByVal value As Boolean)
					_escapeHtmlEntities = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece un valor que indica el nombre de la fuente por defecto.
			''' Los grupos de textos que usen esta fuente no incluirn la propiedad font-face
			''' </summary>
			Public Property DefaultFontName() As String
				Get
					Return _defaultFontName
				End Get
				Set(ByVal value As String)
					_defaultFontName = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece un valor que indica el tamao de fuente que se ignorar.
			''' Los grupos de texto que tengan ese tamao no incluirn la propiedad CSS font-size
			''' </summary>
			Public Property DefaultFontSize() As Integer
				Get
					Return _defaultFontSize
				End Get
				Set(ByVal value As Integer)
					_defaultFontSize = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece un valor que indica si se incrustarn las imgenes en base64
			''' dentro del cdigo HTML  (true), o se guardarn en un archivo (false)
			''' </summary>
			''' <see cref="http://www.sweeting.org/mark/blog/2005/07/12/base64-encoded-images-embedded-in-html"/>
			Public Property IncrustImages() As Boolean
				Get
					Return _incrustImages
				End Get
				Set(ByVal value As Boolean)
					_incrustImages = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece la ruta donde se guardarn las imgenes contenidas en el
			''' cdigo RTF original. Slo se usarar si el valor de IncrustImages es false
			''' </summary>
			Public Property ImagePath() As String
				Get
					Return _imagePath
				End Get
				Set(ByVal value As String)
					_imagePath = value
				End Set
			End Property

			''' <summary>
			''' Obtiene o establece un valor que indica el formato en el que se guardarn las imgenes
			''' incrustadas en el cdigo RTF convertido
			''' </summary>
			Public Property ImageFormat() As ImageFormat
				Get
					Return _imageFormat
				End Get
				Set(ByVal value As ImageFormat)
					_imageFormat = value
				End Set
			End Property

			#End Region

			#Region "Mtodos Pblicos"

			''' <summary>
			''' Convierte una cadena de cdigo RTF a formato HTML
			''' </summary>
			Public Shared Function ConvertCode(ByVal rtf As String) As String
				Dim rtfToHtml As New Rtf2Html()
				Return rtfToHtml.Convert(rtf)
			End Function

			''' <summary>
			''' Convierte una cadena de cdigo RTF a formato HTML
			''' </summary>
			Public Function Convert(ByVal rtf As String) As String
				'Generar arbol DOM
				Dim rtfTree As New RtfTree()
				rtfTree.LoadRtfText(rtf)

				'Inicializar variables empleadas
				_builder = New StringBuilder()
				_htmlFormat = New Format()
				_currentFormat = New Format()
				_fontTable = rtfTree.GetFontTable()
				_colorTable = rtfTree.GetColorTable()

				'Buscar el inicio del contenido visible del documento
				Dim inicio As Integer
				For inicio = 0 To rtfTree.RootNode.FirstChild.ChildNodes.Count - 1
					If rtfTree.RootNode.FirstChild.ChildNodes(inicio).NodeKey = "pard" Then
						Exit For
					End If
				Next inicio

				'Procesar todos los nodos visibles
				ProcessChildNodes(rtfTree.RootNode.FirstChild.ChildNodes, inicio)

				'Cerrar etiquetas pendientes
				_currentFormat.Reset()
				WriteText(String.Empty)

				'Arreglar HTML

				'Arreglar listas
				Dim repairList As New Regex("<span [^>]*></span><span style=""([^""]*)"">(.*?)<br\s+/><" & "/span>", RegexOptions.IgnoreCase Or RegexOptions.Singleline Or RegexOptions.CultureInvariant)

				For Each match As Match In repairList.Matches(_builder.ToString())
					_builder.Replace(match.Value, String.Format("<li style=""{0}"">{1}</li>", match.Groups(1).Value, match.Groups(2).Value))
				Next match

				Dim repairUl As New Regex("(?<!</li>)<li", RegexOptions.IgnoreCase Or RegexOptions.CultureInvariant)

				For Each match As Match In repairUl.Matches(_builder.ToString())
					_builder.Insert(match.Index, "<ul>")
				Next match

				repairUl = New Regex("/li>(?!<li)", RegexOptions.IgnoreCase Or RegexOptions.CultureInvariant)

				For Each match As Match In repairUl.Matches(_builder.ToString())
					_builder.Insert(match.Index + match.Length, "</ul>")
				Next match

				'Generar prrafos (cada 2 <br /><br /> se cambiar por un <p>)
				If AutoParagraph Then
					Dim partes() As String = _builder.ToString().Split(New String() { "<br /><br />" }, StringSplitOptions.RemoveEmptyEntries)
					_builder = New StringBuilder(_builder.Length + 7 * partes.Length)

					For Each parte As String In partes
						_builder.Append("<p>")
						_builder.Append(parte)
						_builder.Append("</p>")
					Next parte
				End If

				Return If(EscapeHtmlEntities, HtmlEntities.Encode(_builder.ToString()), _builder.ToString())
			End Function

			#End Region

			#Region "Mtodos Privados"

			''' <summary>
			''' Funcin encargada de procesar los nodos hijo de un nodo padre RTF un nodo del documento RTF
			''' y generar las etiquetas HTML necesarias
			''' </summary>
			Private Sub ProcessChildNodes(ByVal nodos As RtfNodeCollection, ByVal inicio As Integer)
				For i As Integer = inicio To nodos.Count - 1
					Dim nodo As RtfTreeNode = nodos(i)

					Select Case nodo.NodeType
						Case RtfNodeType.Control

							If nodo.NodeKey = "'" Then 'Smbolos especiales, como tildes y ""
                                WriteText(Encoding.Default.GetString(New Byte() {CByte(nodo.Parameter)}))
							End If

						Case RtfNodeType.Keyword

							Select Case nodo.NodeKey
								Case "pard" 'Reinicio de formato
									_currentFormat.Reset()

								Case "f" 'Tipo de fuente
									If nodo.Parameter < _fontTable.Count Then
										_currentFormat.FontName = _fontTable(nodo.Parameter)
									End If

								Case "cf" 'Color de fuente
									If nodo.Parameter < _colorTable.Count Then
										_currentFormat.ForeColor = _colorTable(nodo.Parameter)
									End If

								Case "highlight" 'Color de fondo
									If nodo.Parameter < _colorTable.Count Then
										_currentFormat.BackColor = _colorTable(nodo.Parameter)
									End If

								Case "fs" 'Tamao de fuente
									_currentFormat.FontSize = nodo.Parameter

								Case "b" 'Negrita
									_currentFormat.Bold = (Not nodo.HasParameter) OrElse nodo.Parameter = 1

								Case "i" 'Cursiva
									_currentFormat.Italic = (Not nodo.HasParameter) OrElse nodo.Parameter = 1

								Case "ul" 'Subrayado ON
									_currentFormat.Underline = True

								Case "ulnone" 'Subrayado OFF
									_currentFormat.Underline = False

								Case "super" 'Superscript
									_currentFormat.Superscript = True
									_currentFormat.Subscript = False

								Case "sub" 'Subindice
									_currentFormat.Subscript = True
									_currentFormat.Superscript = False

								Case "nosupersub"
									_currentFormat.Subscript = False
									_currentFormat.Superscript = _currentFormat.Subscript

								Case "qc" 'Alineacion centrada
									_currentFormat.Alignment = HorizontalAlignment.Center

								Case "qr" 'Alineacion derecha
									_currentFormat.Alignment = HorizontalAlignment.Right

								Case "li" 'tabulacion
									_currentFormat.Margin = nodo.Parameter

								Case "line", "par"
									_builder.Append("<br />")
							End Select


						Case RtfNodeType.Group

							'Procesar nodos hijo, si los tiene
							If nodo.HasChildNodes() Then
								If nodo("pict") IsNot Nothing Then 'El grupo es una imagen
									Dim imageNode As New ImageNode(nodo)
									WriteImage(imageNode)
								Else
									ProcessChildNodes(nodo.ChildNodes, 0)
								End If
							End If

						Case RtfNodeType.Text

							WriteText(nodo.NodeKey)

						Case Else

							Throw New ArgumentOutOfRangeException()
					End Select
				Next i
			End Sub

			''' <summary>
			''' Funcin encargada de aadir texto con el formato actual al cdigo html resultado
			''' </summary>
			Private Sub WriteText(ByVal text As String)
				If _builder.Length > 0 Then
					'Cerrar etiquetas
					If _currentFormat.Bold = False AndAlso _htmlFormat.Bold Then
						_builder.Append("</strong>")
						_htmlFormat.Bold = False
					End If
					If _currentFormat.Italic = False AndAlso _htmlFormat.Italic Then
						_builder.Append("</em>")
						_htmlFormat.Italic = False
					End If
					If _currentFormat.Underline = False AndAlso _htmlFormat.Underline Then
						_builder.Append("</u>")
						_htmlFormat.Underline = False
					End If
					If _currentFormat.Subscript = False AndAlso _htmlFormat.Subscript Then
						_builder.Append("</sub>")
						_htmlFormat.Subscript = False
					End If
					If _currentFormat.Superscript = False AndAlso _htmlFormat.Superscript Then
						_builder.Append("</sup>")
						_htmlFormat.Superscript = False
					End If
					If _currentFormat.CompareFontFormat(_htmlFormat) = False Then 'El formato de fuente ha cambiado
						_builder.Append("</span>")

						'Reiniciar propiedades
						_htmlFormat.Reset()
					End If
				End If

				'Abrir etiquetas necesarias para representar el formato actual
				If _currentFormat.CompareFontFormat(_htmlFormat) = False Then 'El formato de fuente ha cambiado
					Dim estilo As String = String.Empty

					If (Not IgnoreFontNames) AndAlso (Not String.IsNullOrEmpty(_currentFormat.FontName)) AndAlso String.Compare(_currentFormat.FontName, DefaultFontName, True) <> 0 Then
						estilo &= String.Format("font-family:{0};", _currentFormat.FontName)
					End If
					If _currentFormat.FontSize > 0 AndAlso _currentFormat.FontSize \ 2 <> DefaultFontSize Then
						estilo &= String.Format("font-size:{0}pt;", _currentFormat.FontSize \ 2)
					End If
					If _currentFormat.Margin <> _htmlFormat.Margin Then
						estilo &= String.Format("margin-left:{0}px;", _currentFormat.Margin \ 15)
					End If
					If _currentFormat.Alignment <> _htmlFormat.Alignment Then
						estilo &= String.Format("text-align:{0};", _currentFormat.Alignment.ToString().ToLower())
					End If
					If CompareColor(_currentFormat.ForeColor, _htmlFormat.ForeColor) = False Then
						estilo &= String.Format("color:{0};", ColorTranslator.ToHtml(_currentFormat.ForeColor))
					End If
					If CompareColor(_currentFormat.BackColor, _htmlFormat.BackColor) = False Then
						estilo &= String.Format("background-color:{0};", ColorTranslator.ToHtml(_currentFormat.BackColor))
					End If

					_htmlFormat.FontName = _currentFormat.FontName
					_htmlFormat.FontSize = _currentFormat.FontSize
					_htmlFormat.ForeColor = _currentFormat.ForeColor
					_htmlFormat.BackColor = _currentFormat.BackColor
					_htmlFormat.Margin = _currentFormat.Margin
					_htmlFormat.Alignment = _currentFormat.Alignment

					If (Not String.IsNullOrEmpty(estilo)) Then
						_builder.AppendFormat("<span style=""{0}"">", estilo)
					End If
				End If
				If _currentFormat.Superscript AndAlso _htmlFormat.Superscript = False Then
					_builder.Append("<sup>")
					_htmlFormat.Superscript = True
				End If
				If _currentFormat.Subscript AndAlso _htmlFormat.Subscript = False Then
					_builder.Append("<sub>")
					_htmlFormat.Subscript = True
				End If
				If _currentFormat.Underline AndAlso _htmlFormat.Underline = False Then
					_builder.Append("<u>")
					_htmlFormat.Underline = True
				End If
				If _currentFormat.Italic AndAlso _htmlFormat.Italic = False Then
					_builder.Append("<em>")
					_htmlFormat.Italic = True
				End If
				If _currentFormat.Bold AndAlso _htmlFormat.Bold = False Then
					_builder.Append("<strong>")
					_htmlFormat.Bold = True
				End If

				_builder.Append(text.Replace("""", "&quot;").Replace("<", "&lt;").Replace(">", "&gt;"))
			End Sub

			''' <summary>
			''' Funcin encargada de aadir una imagen al cdigo HTML resultado
			''' </summary>
			''' <param name="imageNode"></param>
			Private Sub WriteImage(ByVal imageNode As ImageNode)
				Dim b As Bitmap = imageNode.Bitmap
				Dim imageSize As Size

				If imageNode.DesiredHeight > 0 AndAlso imageNode.DesiredWidth > 0 Then 'Aproximar twips a px dividiendo entre 15
                    imageSize = New Size(imageNode.DesiredWidth \ 15, imageNode.DesiredHeight \ 15)
				Else
					imageSize = b.Size
				End If

				'Buscar codec
				Dim imageCodec As ImageCodecInfo = Nothing
				For Each codec As ImageCodecInfo In ImageCodecInfo.GetImageDecoders()
					If imageCodec Is Nothing OrElse codec.FormatID = ImageFormat.Guid Then
						imageCodec = codec
					End If
				Next codec

				'Reducir imagen si el tamao final es menor que el actual
				If b.Size.Height > imageSize.Height OrElse b.Size.Width > imageSize.Width Then
					Dim bmpResized As New Bitmap(imageSize.Width, imageSize.Height)
					Dim g As Graphics = Graphics.FromImage(bmpResized)
					g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic
					g.DrawImage(b, 0, 0, imageSize.Width, imageSize.Height)
					b.Dispose()
					b = bmpResized
				End If

				'Generar imagen en el formato adecuado
				Dim imageSrc As String
				If IncrustImages Then
                    Using buffer As New MemoryStream()
                        b.Save(buffer, ImageFormat)

                        imageSrc = String.Format("data:{0};base64,{1}", imageCodec.MimeType, System.Convert.ToBase64String(buffer.GetBuffer(), 0, CInt(Fix(buffer.Length))))
                    End Using
				Else
					Dim index As Integer = 1
					Dim ext As String = imageCodec.FilenameExtension.Split(";"c)(0).Substring(2).ToLower()
                    Dim mPath As String

                    Do
                        mPath = Path.Combine(ImagePath, String.Format("Imagen{0}.{1}", index, ext))
                        index += 1
                    Loop While File.Exists(mPath)

                    b.Save(mPath, ImageFormat)

                    'imageSrc = "file://" + path;
                    imageSrc = mPath
				End If

				'Aadir imagen al resultado HTML
				_builder.Append(String.Format("<img src=""{0}"" style=""width:{1}px;height:{2}px;"" />", imageSrc, imageSize.Width, imageSize.Height))

				'Limpiar memoria
				b.Dispose()
			End Sub

			''' <summary>
			''' Compara dos colores sin tener en cuenta el canal alfa
			''' </summary>
			Private Shared Function CompareColor(ByVal a As Color, ByVal b As Color) As Boolean
				Return a.R = b.R AndAlso a.G = b.G AndAlso a.B = b.B
			End Function

			#End Region

			#Region "Nested type: Format"

			''' <summary>
			''' Representa el formato que puede tomar un conjunto de texto
			''' </summary>
			Private Class Format
				Public Italic As Boolean
				Public Bold As Boolean
				Public Subscript As Boolean
				Public Underline As Boolean
				Public Superscript As Boolean

				Public FontName As String
				Public FontSize As Integer
				Public ForeColor As Color
				Public BackColor As Color
				Public Margin As Integer
				Public Alignment As HorizontalAlignment

				Public Sub New()
					Reset()
				End Sub

				''' <summary>
				''' Compara las propiedades FontName, FontSize, Margin, ForeColor, BackColor y Alignment con otro
				''' objeto de esta clase
				''' </summary>
				Public Function CompareFontFormat(ByVal format As Format) As Boolean
					Return String.Compare(FontName, format.FontName, True) = 0 AndAlso FontSize = format.FontSize AndAlso ForeColor = format.ForeColor AndAlso BackColor = format.BackColor AndAlso Margin = format.Margin AndAlso Alignment = format.Alignment
				End Function

				Public Sub Reset()
					FontName = String.Empty
					FontSize = 0
					ForeColor = Color.Black
					BackColor = Color.White
					Margin = 0
					Alignment = HorizontalAlignment.Left
				End Sub
			End Class

			Private Enum HorizontalAlignment
				Left
				Right
				Center
			End Enum

			#End Region
		End Class
	End Namespace
End Namespace