'************************************************* ******************************
'* This file is part of NRtfTree Library.
'*
'* NRtfTree Library is free software; you can redistribute it and / or modify
'* It under the terms of the GNU Lesser General Public License as published by
'* The Free Software Foundation, either version 3 of the License, or
'* (At your option) any later version.
'*
'* NRtfTree Library is distributed in the hope that it will be useful,
'* But WITHOUT ANY WARRANTY; without even the implied warranty of
'* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
'* GNU Lesser General Public License for more details.
'*
'* You should have received a copy of the GNU Lesser General Public License
'* Along with this program. If not, see <http://www.gnu.org/licenses/> .
'************************************************* ******************************

'************************************************* ******************************
'* Library: NRtfTree
'* Version: v0.3.0
'* Date: 01/05/2009
'* Copyright: 2006-2009 Salvador Gomez
'* E-mail: sgoliver.net @ gmail.com
'* Home Page: http://www.sgoliver.net
'* SF Project: http://nrtftree.sourceforge.net
'* Http://sourceforge.net/projects/nrtftree
'* Class: ImageNode
'* Description: RTF node that contains the information specialist in an image.
'* VB.NET Version Contributed by: Terence L. Wallace
'* ************************************************ *****************************


Imports Microsoft.VisualBasic
Imports System
Imports System.Text
Imports Net.Sgoliver.NRtfTree.Core
Imports System.IO
Imports System.Globalization
Imports System.Drawing
Imports System.Drawing.Imaging

Namespace Net.Sgoliver.NRtfTree
    Namespace Util
        ''' <summary>
        ''' Encapsulates a node of type RTF Image (Keyword "\ pict")
        ''' </summary>
        Public Class ImageNode
            Inherits Net.Sgoliver.NRtfTree.Core.RtfTreeNode

            ''' <summary>
            ''' Array of bytes with information of the image.
            ''' </summary>
            Private data() As Byte

#Region "Constructors"

            ''' <summary>
            ''' ImageNode class constructor.
            ''' </summary>
            ''' <param name="node"> Node obtendrn RTF from which the image data. </param>
            Public Sub New(ByVal node As RtfTreeNode)
                If node IsNot Nothing Then
                    'Assign all fields in the node
                    Me.NodeKey = node.NodeKey
                    Me.HasParameter = node.HasParameter
                    Me.Parameter = node.Parameter
                    Me.ParentNode = node.ParentNode
                    Me.RootNode = node.RootNode
                    Me.NodeType = node.NodeType

                    Me.ChildNodes = New RtfNodeCollection()
                    Me.ChildNodes.AddRange(node.ChildNodes)

                    'Get the image data as an array of bytes
                    getImageData()
                End If
            End Sub

#End Region

#Region "Properties"

            ''' <summary>
            ''' Returns a string with the contents of the image in hexadecimal.
            ''' </summary>
            Public ReadOnly Property HexData() As String
                Get
                    Return Me.SelectSingleChildNode(RtfNodeType.Text).NodeKey
                End Get
            End Property

            ''' <summary>
            ''' Returns the original image format.
            ''' </summary>
            Public ReadOnly Property ImageFormat() As System.Drawing.Imaging.ImageFormat
                Get
                    If Me.SelectSingleChildNode("jpegblip") IsNot Nothing Then
                        Return System.Drawing.Imaging.ImageFormat.Jpeg
                    ElseIf Me.SelectSingleChildNode("pngblip") IsNot Nothing Then
                        Return System.Drawing.Imaging.ImageFormat.Png
                    ElseIf Me.SelectSingleChildNode("emfblip") IsNot Nothing Then
                        Return System.Drawing.Imaging.ImageFormat.Emf
                    ElseIf Me.SelectSingleChildNode("wmetafile") IsNot Nothing Then
                        Return System.Drawing.Imaging.ImageFormat.Wmf
                    ElseIf Me.SelectSingleChildNode("dibitmap") IsNot Nothing OrElse Me.SelectSingleChildNode("wbitmap") IsNot Nothing Then
                        Return System.Drawing.Imaging.ImageFormat.Bmp
                    Else
                        Return Nothing
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the image width (in twips).
            ''' </summary>
            Public ReadOnly Property Width() As Integer
                Get
                    Dim node As RtfTreeNode = Me.SelectSingleChildNode("picw")

                    If node IsNot Nothing Then
                        Return node.Parameter
                    Else
                        Return -1
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the height of the image (in twips).
            ''' </summary>
            Public ReadOnly Property Height() As Integer
                Get
                    Dim node As RtfTreeNode = Me.SelectSingleChildNode("pich")

                    If node IsNot Nothing Then
                        Return node.Parameter
                    Else
                        Return -1
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the width of the image target (in twips).
            ''' </summary>
            Public ReadOnly Property DesiredWidth() As Integer
                Get
                    Dim node As RtfTreeNode = Me.SelectSingleChildNode("picwgoal")

                    If node IsNot Nothing Then
                        Return node.Parameter
                    Else
                        Return -1
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the highest goal of the image (in twips).
            ''' </summary>
            Public ReadOnly Property DesiredHeight() As Integer
                Get
                    Dim node As RtfTreeNode = Me.SelectSingleChildNode("pichgoal")

                    If node IsNot Nothing Then
                        Return node.Parameter
                    Else
                        Return -1
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the horizontal scale of the image, as a percentage.
            ''' </summary>
            Public ReadOnly Property ScaleX() As Integer
                Get
                    Dim node As RtfTreeNode = Me.SelectSingleChildNode("picscalex")

                    If node IsNot Nothing Then
                        Return node.Parameter
                    Else
                        Return -1
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the vertical scale of the image, as a percentage.
            ''' </summary>
            Public ReadOnly Property ScaleY() As Integer
                Get
                    Dim node As RtfTreeNode = Me.SelectSingleChildNode("picscaley")

                    If node IsNot Nothing Then
                        Return node.Parameter
                    Else
                        Return -1
                    End If
                End Get
            End Property

            ''' <summary>
            ''' Returns the image to a bitmap object.
            ''' </summary>
            Public ReadOnly Property Bitmap() As Bitmap
                Get
                    Dim stream As New MemoryStream(Me.GetByteData(), 0, data.Length)
                    Return New Bitmap(stream)
                End Get
            End Property

#End Region

#Region "Public Methods"

            ''' <summary>
            ''' Returns an array of bytes with the contents of the image.
            ''' </summary>
            ''' <return> Array of bytes to the image content. </return>
            Public Function GetByteData() As Byte()
                Return data
            End Function

            ''' <summary>
            ''' Save an image file with the original format.
            ''' </summary>
            ''' <param name="filePath"> File path where you save the image. </param>
            Public Sub SaveImage(ByVal filePath As String)
                If data IsNot Nothing Then
                    Dim stream As New MemoryStream(Me.GetByteData(), 0, data.Length)

                    'Write to file any type of image
                    Dim bitmap As New Bitmap(stream)
                    bitmap.Save(filePath, Me.ImageFormat)
                End If
            End Sub

            ''' <summary>
            ''' Save an image file with a specific format indicated as parmetro.
            ''' </summary>
            ''' <param name="filePath"> File path where you save the image. </param>
            ''' <param name="format"> Format in which to write the image. </param>
            Public Sub SaveImage(ByVal filePath As String, ByVal format As System.Drawing.Imaging.ImageFormat)
                If data IsNot Nothing Then
                    Dim stream As New MemoryStream(data, 0, data.Length)

                    'System.Drawing.Imaging.Metafile metafile = new System.Drawing.Imaging.Metafile (stream);

                    'Write the byte array directly to a file. "Jpg"
                    'FileStream fs = new FileStream ( "c: \ \ prueba.jpg" FileMode.CreateNew);
                    'BinaryWriter w = new BinaryWriter (fs);
                    'w.write (image, 0, imageSize);
                    'w.close ();
                    'fs.Close ();

                    'Write to file any type of image
                    Dim bitmap As New Bitmap(stream)
                    bitmap.Save(filePath, format)
                End If
            End Sub

#End Region

#Region "Private Methods"

            ''' <summary>
            ''' Gets the image data from the information contained in the node RTF.
            ''' </summary>
            Private Sub getImageData()
                'Format 1 (Word 97-2000): (\ * \ shppict (\ pict \ jpegblip <datos> )) (\ nonshppict (\ pict \ wmetafile8 <datos> ))
                'Format 2 (Wordpad): (\ pict \ wmetafile8 <datos> )

                Dim Text As String = ""

                If Me.FirstChild.NodeKey = "pict" Then
                    Text = Me.SelectSingleChildNode(RtfNodeType.Text).NodeKey

                    Dim dataSize As Integer = Text.Length \ 2
                    data = New Byte(dataSize - 1) {}

                    Dim sbaux As New StringBuilder(2)

                    For i As Integer = 0 To Text.Length - 1
                        sbaux.Append(Text.Chars(i))

                        If sbaux.Length = 2 Then
                            data(i \ 2) = Byte.Parse(sbaux.ToString(), NumberStyles.HexNumber)
                            sbaux.Remove(0, 2)
                        End If
                    Next i
                End If
            End Sub

#End Region
        End Class
    End Namespace
End Namespace
