'************************************************* ******************************
'* This file is part of NRtfTree Library.
'*
'* NRtfTree Library is free software; you can redistribute it and / or modify
'* It under the terms of the GNU Lesser General Public License as published by
'* The Free Software Foundation, either version 3 of the License, or
'* (At your option) any later version.
'*
'* NRtfTree Library is distributed in the hope that it will be useful,
'* But WITHOUT ANY WARRANTY; without even the implied warranty of
'* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
'* GNU Lesser General Public License for more details.
'*
'* You should have received a copy of the GNU Lesser General Public License
'* Along with this program. If not, see <http://www.gnu.org/licenses/> .
'************************************************* ******************************

'************************************************* ******************************
'* Library: NRtfTree
'* Version: v0.3.0
'* Date: 01/05/2009
'* Copyright: 2006-2009 Salvador Gomez
'* E-mail: sgoliver.net @ gmail.com
'* Home Page: http://www.sgoliver.net
'* SF Project: http://nrtftree.sourceforge.net
'* Http://sourceforge.net/projects/nrtftree
'* Class: RtfDocument
'* Description: Class for Generacin RTF documents.
'* VB.NET Version Contributed by: Terence L. Wallace
'* ************************************************ *****************************


Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Text
Imports System.Drawing
Imports System.Globalization
Imports System.IO
Imports Net.Sgoliver.NRtfTree.Core

Namespace Net.Sgoliver.NRtfTree
    Namespace Util
        ''' <summary>
        ''' Generic class for RTF documents.
        ''' </summary> 
        Public Class RtfDocument
#Region "Attributes private"

            ''' <summary>
            ''' Path of file to generate.
            ''' </summary>
            Private path As String

            ''' <summary>
            ''' Encoding of the document.
            ''' </summary>
            Private encoding As Encoding

            ''' <summary>
            ''' Table of fonts in the document.
            ''' </summary>
            Private fontTable As RtfFontTable

            ''' <summary>
            ''' Color table the document.
            ''' </summary>
            Private colorTable As RtfColorTable

            ''' <summary>
            ''' RTF document tree.
            ''' </summary>
            Private tree As RtfTree

            ''' <summary>
            ''' Group of the document.
            ''' </summary>
            Private mainGroup As RtfTreeNode

            ''' <summary>
            ''' Current format of character.
            ''' </summary>
            Private currentFormat As RtfCharFormat

            ''' <summary>
            ''' Subparagraph current format.
            ''' </summary>
            Private currentParFormat As RtfParFormat

            ''' <summary>
            ''' Document format.
            ''' </summary>
            Private docFormat As RtfDocumentFormat

#End Region

#Region "Constructors"

            ''' <summary>
            ''' RtfDocument class constructor.
            ''' </summary>
            ''' <param name="path"> Path of file to generate. </param>
            ''' <param name="enc"> Encoding of the document to generate. </param>
            Public Sub New(ByVal path As String, ByVal enc As Encoding)
                Me.path = path
                Me.encoding = enc

                fontTable = New RtfFontTable()
                fontTable.AddFont("Arial") 'Default font

                colorTable = New RtfColorTable()
                colorTable.AddColor(Color.Black) 'Default color

                currentFormat = Nothing
                currentParFormat = New RtfParFormat()
                docFormat = New RtfDocumentFormat()

                tree = New RtfTree()
                mainGroup = New RtfTreeNode(RtfNodeType.Group)

                InitializeTree()
            End Sub

            ''' <summary>
            ''' RtfDocument class constructor. Encoding is to use the system default.
            ''' </summary>
            ''' <param name="path"> Path of file to generate. </param>
            Public Sub New(ByVal path As String)
                Me.New(path, encoding.Default)

            End Sub

#End Region

#Region "Public Methods"

            ''' <summary>
            ''' Closes the RTF document.
            ''' </summary>
            Public Sub Close()
                InsertFontTable()
                InsertColorTable()
                InsertGenerator()
                InsertDocSettings()

                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "par", False, 0))
                tree.RootNode.AppendChild(mainGroup)

                tree.SaveRtf(path)
            End Sub

            ''' <summary>
            ''' Insert a piece of text in the document in a text format specified.
            ''' </summary>
            ''' <param name="text"> Text to be inserted. </param>
            ''' <param name="format"> Format the text to insert. </param>
            Public Sub AddText(ByVal text As String, ByVal format As RtfCharFormat)
                UpdateFontTable(format)
                UpdateColorTable(format)

                UpdateCharFormat(format)

                InsertText(text)
            End Sub

            ''' <summary>
            ''' Insert a piece of text in the document with the current text format.
            ''' </summary>
            ''' <param name="text"> Text to be inserted. </param>
            Public Sub AddText(ByVal text As String)
                InsertText(text)
            End Sub

            ''' <summary>
            ''' Inserts a given NUMBER of line breaks in the document.
            ''' </summary>
            ''' <param name="n"> Numbers of line breaks inserted. </param>
            Public Sub AddNewLine(ByVal n As Integer)
                For i As Integer = 0 To n - 1
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "line", False, 0))
                Next i
            End Sub

            ''' <summary>
            ''' Insert a jump of line in the document.
            ''' </summary>
            Public Sub AddNewLine()
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "line", False, 0))
            End Sub

            ''' <summary>
            ''' Start a new subparagraph.
            ''' </summary>
            Public Sub AddNewParagraph()
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "par", False, 0))
            End Sub

            ''' <summary>
            ''' Insert a jump NUMBER subparagraph given in the paper.
            ''' </summary>
            ''' <param name="n"> Numbers subparagraph breaks inserted. </param>
            Public Sub AddNewParagraph(ByVal n As Integer)
                For i As Integer = 0 To n - 1
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "par", False, 0))
                Next i
            End Sub

            ''' <summary>
            ''' Start a new subparagraph to the specified format.
            ''' </summary>
            Public Sub AddNewParagraph(ByVal format As RtfParFormat)
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "par", False, 0))

                UpdateParFormat(format)
            End Sub

            ''' <summary>
            ''' Insert an image into the document.
            ''' </summary>
            ''' <param name="path"> Image Path inserted. </param>
            ''' <param name="width"> Desired width of the image in the document. </param>
            ''' <param name="height"> High expectation of the image in the document. </param>
            Public Sub AddImage(ByVal path As String, ByVal width As Integer, ByVal height As Integer)
                Dim fStream As FileStream = Nothing
                Dim br As BinaryReader = Nothing

                Try
                    Dim data() As Byte = Nothing

                    Dim fInfo As New FileInfo(path)
                    Dim numBytes As Long = fInfo.Length

                    fStream = New FileStream(path, FileMode.Open, FileAccess.Read)
                    br = New BinaryReader(fStream)

                    data = br.ReadBytes(CInt(Fix(numBytes)))

                    Dim hexdata As New StringBuilder()

                    For i As Integer = 0 To data.Length - 1
                        hexdata.Append(GetHexa(data(i)))
                    Next i

                    Dim img As Image = Image.FromFile(path)

                    Dim imgGroup As New RtfTreeNode(RtfNodeType.Group)
                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "pict", False, 0))

                    Dim format As String = ""
                    If path.ToLower().EndsWith("wmf") Then
                        format = "emfblip"
                    Else
                        format = "jpegblip"
                    End If

                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, format, False, 0))


                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "picw", True, img.Width * 20))
                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "pich", True, img.Height * 20))
                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "picwgoal", True, width * 20))
                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "pichgoal", True, height * 20))
                    imgGroup.AppendChild(New RtfTreeNode(RtfNodeType.Text, hexdata.ToString(), False, 0))

                    mainGroup.AppendChild(imgGroup)
                Finally
                    br.Close()
                    fStream.Close()
                End Try
            End Sub

            ''' <summary>
            ''' Sets the default character format.
            ''' </summary>
            Public Sub ResetCharFormat()
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "plain", False, 0))
            End Sub

            ''' <summary>
            ''' Sets the default format subparagraph.
            ''' </summary>
            Public Sub ResetParFormat()
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "pard", False, 0))
            End Sub

            ''' <summary>
            ''' Sets the character format yprrafo default.
            ''' </summary>
            Public Sub ResetFormat()
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "pard", False, 0))
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "plain", False, 0))
            End Sub

            ''' <summary>
            ''' Updates the values of the properties of document format.
            ''' </summary>
            ''' <param name="format"> Document Format. </param>
            Public Sub UpdateDocFormat(ByVal format As RtfDocumentFormat)
                docFormat.MarginL = format.MarginL
                docFormat.MarginR = format.MarginR
                docFormat.MarginT = format.MarginT
                docFormat.MarginB = format.MarginB
            End Sub

            ''' <summary>
            ''' Updates the property values yprrafo text format.
            ''' </summary>
            ''' <param name="format"> Format text to be inserted. </param>
            Public Sub UpdateCharFormat(ByVal format As RtfCharFormat)
                If currentFormat IsNot Nothing Then
                    SetFormatColor(format.Color)
                    SetFormatSize(format.Size)
                    SetFormatFont(format.Font)

                    SetFormatBold(format.Bold)
                    SetFormatItalic(format.Italic)
                    SetFormatUnderline(format.Underline)
                Else 'currentFormat == null
                    Dim indColor As Integer = colorTable.IndexOf(format.Color)

                    If indColor = -1 Then
                        colorTable.AddColor(format.Color)
                        indColor = colorTable.IndexOf(format.Color)
                    End If

                    Dim indFont As Integer = fontTable.IndexOf(format.Font)

                    If indFont = -1 Then
                        fontTable.AddFont(format.Font)
                        indFont = fontTable.IndexOf(format.Font)
                    End If

                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "cf", True, indColor))
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "fs", True, format.Size * 2))
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "f", True, indFont))

                    If format.Bold Then
                        mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "b", False, 0))
                    End If

                    If format.Italic Then
                        mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "i", False, 0))
                    End If

                    If format.Underline Then
                        mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "ul", False, 0))
                    End If

                    currentFormat = New RtfCharFormat()
                    currentFormat.Color = format.Color
                    currentFormat.Size = format.Size
                    currentFormat.Font = format.Font
                    currentFormat.Bold = format.Bold
                    currentFormat.Italic = format.Italic
                    currentFormat.Underline = format.Underline
                End If
            End Sub

            ''' <summary>
            ''' Sets the format of past and parmetro subparagraph.
            ''' </summary>
            ''' <param name="format"> Subparagraph format to use. </param>
            Public Sub UpdateParFormat(ByVal format As RtfParFormat)
                SetAlignment(format.Alignment)
                SetLeftIndentation(format.LeftIndentation)
                SetRightIndentation(format.RightIndentation)
            End Sub

            ''' <summary>
            ''' Alineacin stable of text within the subparagraph.
            ''' </summary>
            ''' <param name="align"> Alineacin type. </param>
            Public Sub SetAlignment(ByVal align As TextAlignment)
                If currentParFormat.Alignment <> align Then
                    Dim keyword As String = ""

                    Select Case align
                        Case TextAlignment.Left
                            keyword = "ql"
                        Case TextAlignment.Right
                            keyword = "qr"
                        Case TextAlignment.Centered
                            keyword = "qc"
                        Case TextAlignment.Justified
                            keyword = "qj"
                    End Select

                    currentParFormat.Alignment = align
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, keyword, False, 0))
                End If
            End Sub

            ''' <summary>
            ''' Sets the left of subparagraph bleeds.
            ''' </summary>
            ''' <param name="val"> Sangra left in centimeters. </param>
            Public Sub SetLeftIndentation(ByVal val As Single)
                If currentParFormat.LeftIndentation <> val Then
                    currentParFormat.LeftIndentation = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "li", True, calcTwips(val)))
                End If
            End Sub

            ''' <summary>
            ''' Establishes the right of subparagraph bleeds.
            ''' </summary>
            ''' <param name="val"> Bleeds right into centimeters. </param>
            Public Sub SetRightIndentation(ByVal val As Single)
                If currentParFormat.RightIndentation <> val Then
                    currentParFormat.RightIndentation = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "ri", True, calcTwips(val)))
                End If
            End Sub

            ''' <summary>
            ''' Sets the bold font.
            ''' </summary>
            ''' <param name="val"> Indicative bold font. </param>
            Public Sub SetFormatBold(ByVal val As Boolean)
                If currentFormat.Bold <> val Then
                    currentFormat.Bold = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "b", If(val, False, True), 0))
                End If
            End Sub

            ''' <summary>
            ''' Sets the italic font.
            ''' </summary>
            ''' <param name="val"> Indicative italic font. </param>
            Public Sub SetFormatItalic(ByVal val As Boolean)
                If currentFormat.Italic <> val Then
                    currentFormat.Italic = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "i", If(val, False, True), 0))
                End If
            End Sub

            ''' <summary>
            ''' Sets the source underlined.
            ''' </summary>
            ''' <param name="val"> Indicative underlined font. </param>
            Public Sub SetFormatUnderline(ByVal val As Boolean)
                If currentFormat.Underline <> val Then
                    currentFormat.Underline = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "ul", If(val, False, True), 0))
                End If
            End Sub

            ''' <summary>
            ''' Sets the current font color.
            ''' </summary>
            ''' <param name="val"> Font color. </param>
            Public Sub SetFormatColor(ByVal val As Color)
                If currentFormat.Color.ToArgb() <> val.ToArgb() Then
                    Dim indColor As Integer = colorTable.IndexOf(val)

                    If indColor = -1 Then
                        colorTable.AddColor(val)
                        indColor = colorTable.IndexOf(val)
                    End If

                    currentFormat.Color = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "cf", True, indColor))
                End If
            End Sub

            ''' <summary>
            ''' Dick size Sets the current font.
            ''' </summary>
            ''' <param name="val"> Tamao from the source. </param>
            Public Sub SetFormatSize(ByVal val As Integer)
                If currentFormat.Size <> val Then
                    currentFormat.Size = val

                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "fs", True, val * 2))
                End If
            End Sub

            ''' <summary>
            ''' Sets the current font.
            ''' </summary>
            ''' <param name="val"> Font. </param>
            Public Sub SetFormatFont(ByVal val As String)
                If currentFormat.Font <> val Then
                    Dim indFont As Integer = fontTable.IndexOf(val)

                    If indFont = -1 Then
                        fontTable.AddFont(val)
                        indFont = fontTable.IndexOf(val)
                    End If

                    currentFormat.Font = val
                    mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "f", True, indFont))
                End If
            End Sub

#End Region

#Region "Private Methods"

            ''' <summary>
            ''' Gets the code of a hexadecimal integer.
            ''' </summary>
            ''' <param name="code"> NUMBER integer. </param>
            ''' <returns> Cdigo hexadecimal integer passed as parmetro. </returns>
            Private Function GetHexa(ByVal code As Byte) As String
                Dim hexa As String = Convert.ToString(code, 16)

                If hexa.Length = 1 Then
                    hexa = "0" & hexa
                End If

                Return hexa
            End Function

            ''' <summary>
            ''' Insert the code of the RTF font table in the document.
            ''' </summary>
            Private Sub InsertFontTable()
                Dim ftGroup As New RtfTreeNode(RtfNodeType.Group)

                ftGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "fonttbl", False, 0))

                For i As Integer = 0 To fontTable.Count - 1
                    Dim ftFont As New RtfTreeNode(RtfNodeType.Group)
                    ftFont.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "f", True, i))
                    ftFont.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "fnil", False, 0))
                    ftFont.AppendChild(New RtfTreeNode(RtfNodeType.Text, fontTable(i) & ";", False, 0))

                    ftGroup.AppendChild(ftFont)
                Next i

                mainGroup.InsertChild(5, ftGroup)
            End Sub

            ''' <summary>
            ''' Insert the code RTF color table in the document.
            ''' </summary>
            Private Sub InsertColorTable()
                Dim ctGroup As New RtfTreeNode(RtfNodeType.Group)

                ctGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "colortbl", False, 0))

                For i As Integer = 0 To colorTable.Count - 1
                    ctGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "red", True, colorTable(i).R))
                    ctGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "green", True, colorTable(i).G))
                    ctGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "blue", True, colorTable(i).B))
                    ctGroup.AppendChild(New RtfTreeNode(RtfNodeType.Text, ";", False, 0))
                Next i

                mainGroup.InsertChild(6, ctGroup)
            End Sub

            ''' <summary>
            ''' Insert the code of the application generating RTF document.
            ''' </summary>
            Private Sub InsertGenerator()
                Dim genGroup As New RtfTreeNode(RtfNodeType.Group)

                genGroup.AppendChild(New RtfTreeNode(RtfNodeType.Control, "*", False, 0))
                genGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "generator", False, 0))
                genGroup.AppendChild(New RtfTreeNode(RtfNodeType.Text, "NRtfTree Library 1.3.0;", False, 0))

                mainGroup.InsertChild(7, genGroup)
            End Sub

            ''' <summary>
            ''' Inserts all text nodes and control needed to represent a particular text.
            ''' </summary>
            ''' <param name="text"> Text to be inserted. </param>
            Private Sub InsertText(ByVal text As String)
                Dim i As Integer = 0
                Dim code As Integer = 0

                Do While i < text.Length
                    code = Char.ConvertToUtf32(text, i)

                    If code >= 32 AndAlso code < 128 Then
                        Dim s As New StringBuilder("")

                        Do While i < text.Length AndAlso code >= 32 AndAlso code < 128
                            s.Append(text.Chars(i))

                            i += 1

                            If i < text.Length Then
                                code = Char.ConvertToUtf32(text, i)
                            End If
                        Loop

                        mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Text, s.ToString(), False, 0))
                    Else
                        Dim bytes() As Byte = encoding.GetBytes(New Char() {text.Chars(i)})

                        mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Control, "'", True, bytes(0)))

                        i += 1
                    End If
                Loop
            End Sub

            ''' <summary>
            ''' Updates the table with a new supply sources if needed.
            ''' </summary>
            ''' <param name="format"></param>
            Private Sub UpdateFontTable(ByVal format As RtfCharFormat)
                If fontTable.IndexOf(format.Font) = -1 Then
                    fontTable.AddFont(format.Font)
                End If
            End Sub

            ''' <summary>
            ''' Updates the color table with a new color if necessary.
            ''' </summary>
            ''' <param name="format"></param>
            Private Sub UpdateColorTable(ByVal format As RtfCharFormat)
                If colorTable.IndexOf(format.Color) = -1 Then
                    colorTable.AddColor(format.Color)
                End If
            End Sub

            ''' <summary>
            ''' Initializes the RTF tree with all the keys in the document header.
            ''' </summary>
            Private Sub InitializeTree()
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "rtf", True, 1))
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "ansi", False, 0))
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "ansicpg", True, encoding.CodePage))
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "deff", True, 0))
                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "deflang", True, CultureInfo.CurrentCulture.LCID))

                mainGroup.AppendChild(New RtfTreeNode(RtfNodeType.Keyword, "pard", False, 0))
            End Sub

            ''' <summary>
            ''' Insert the document formatting properties
            ''' </summary>
            Private Sub InsertDocSettings()
                Dim indInicioTexto As Integer = mainGroup.ChildNodes.IndexOf("pard")

                'Generic Properties

                mainGroup.InsertChild(indInicioTexto, New RtfTreeNode(RtfNodeType.Keyword, "viewkind", True, 4))
                mainGroup.InsertChild(indInicioTexto, New RtfTreeNode(RtfNodeType.Keyword, "uc", True, 1))
                indInicioTexto += 1

                'RtfDocumentFormat Properties

                mainGroup.InsertChild(indInicioTexto, New RtfTreeNode(RtfNodeType.Keyword, "margl", True, calcTwips(docFormat.MarginL)))
                indInicioTexto += 1
                mainGroup.InsertChild(indInicioTexto, New RtfTreeNode(RtfNodeType.Keyword, "margr", True, calcTwips(docFormat.MarginR)))
                indInicioTexto += 1
                mainGroup.InsertChild(indInicioTexto, New RtfTreeNode(RtfNodeType.Keyword, "margt", True, calcTwips(docFormat.MarginT)))
                indInicioTexto += 1
                mainGroup.InsertChild(indInicioTexto, New RtfTreeNode(RtfNodeType.Keyword, "margb", True, calcTwips(docFormat.MarginB)))
                indInicioTexto += 1
            End Sub

            ''' <summary>
            ''' Convert between centimeters and twips.
            ''' </summary>
            ''' <param name="centimeters"> Value in centimeters. </param>
            ''' <returns> Value in twips. </returns>
            Private Function calcTwips(ByVal centimeters As Single) As Integer
                '1 Inches = 2.54 centimeters
                '1 Inches = 1440 twips
                '20 Twip = 1 pixel

                'X centimeters -> (X * 1440) / 2.54 twips

                Return CInt(Fix((centimeters * 1440.0F) / 2.54F))
            End Function

#End Region
        End Class
    End Namespace
End Namespace
